#ifndef __SEZNAM_H__
#define __SEZNAM_H__
#include <iostream>
#include <type_traits>
using namespace std;

template<typename T> 
class seznam
{
	class prvek;
	prvek *hlava, *zarazka;
	mutable prvek *aktualni;
	void tunel(seznam &s);
	static_assert(is_trivially_copyable<T>::value, "Typ ukladanych dat musi byt trivialne kopirovatelny");
public:
	typedef T typ;
	seznam();
	~seznam();
	bool prazdny() const;
	void vyprazdni();
	seznam(const seznam &);
	seznam(seznam&&);
	seznam(initializer_list<T> hodnoty);
	seznam &operator=(const seznam &s);
	seznam &operator=(seznam &&s);
	seznam & seznam::operator+=(const seznam & s);
	seznam operator+(const seznam &s) const;
	void vloz(T co);
	void vloz_na_konec(T co);
	bool najdi(T co) const;
	T &aktualni_hodnota() const;
	bool dalsi() const;
	bool posledni() const;
	void reset() const;
	void na_konec() const;
};

template<typename T>
bool seznam<T>::dalsi() const
{
	if (prazdny() || posledni())
	{
		return false;
	}
	aktualni = aktualni->dalsi();
	return true;
}

template<typename T>
void seznam<T>::tunel(seznam<T>& s)
{
	swap(hlava, s.hlava);
	swap(zarazka, s.zarazka);
	aktualni = hlava;
}


template<typename T>
T &seznam<T>::aktualni_hodnota() const
{
	if (prazdny())
	{
		throw out_of_range("prazdny seznam");
	}
	return aktualni->data();
}

template<typename T>
void seznam<T>::vloz(T co)
{
	if (prazdny())
	{
		hlava = new prvek(co, hlava);
		aktualni = hlava;
	}
	else
	{
		prvek *pom = new prvek(co, aktualni->dalsi());
		aktualni->dalsi(pom);
		aktualni = pom;
	}
}

template<typename T>
bool seznam<T>::najdi(T co) const
{
	for (prvek *pom = hlava; pom != zarazka; pom = pom->dalsi())
	{
		if (pom->data() == co)
		{
			aktualni = pom;
			return true;
		}
	}
	return false;
}

template<typename T>
seznam<T> seznam<T>::operator+(const seznam<T> &s) const
{
	seznam<T> vysledek;
	vysledek += *this;
	vysledek += s;
	return vysledek;
}



template<typename T>
bool seznam<T>::posledni() const
{
	if (prazdny())
	{
		return false;
	}
	return aktualni->dalsi() == zarazka;
}

template<typename T>
void seznam<T>::reset() const
{
	if (!prazdny())
	{
		aktualni = hlava;
	}
}

template<typename T>
void seznam<T>::na_konec() const
{
	if (!prazdny())
	{
		while (dalsi())
			;
	}
}



template<typename T>
seznam<T>::seznam()
{
	hlava = zarazka = new prvek(T(), nullptr);
	aktualni = zarazka;
	cout << "konstruktor () " <<  endl;
}

template<typename T>
seznam<T>::seznam(const seznam<T> &s)
	: seznam<T>()
{
	if (!s.prazdny())
	{
		operator+=(s);
		reset();
	}
	cout << "konstruktor (&)" << endl;
}

template<typename T>
seznam<T>::seznam(seznam<T> && s)
	: seznam<T>()
{
	tunel(s);
	cout << "konstruktor (&&)" << endl;
}


template<typename T>
seznam<T>::seznam(initializer_list<T> hodnoty)
	:seznam<T>()
{
	for (T pom : hodnoty)
	{
		vloz_na_konec(pom);
	}
	reset();
	cout << "konstruktor {}" << endl;
}


template<typename T>
seznam<T> &seznam<T>::operator+=(const seznam<T> & s)
{
	if (!s.prazdny())
	{
		prvek *pom = s.aktualni;
		s.reset();
		do
		{
			vloz_na_konec(s.aktualni_hodnota());
		} while (s.dalsi());
		s.aktualni = pom;
	}
	reset();
	return *this;
}

template<typename T>
void seznam<T>::vloz_na_konec(T co)
{
	zarazka->data() = co;
	aktualni = zarazka;
	prvek *pom = new prvek(T());
	zarazka->dalsi(pom);
	zarazka = pom;
}

template<typename T>
seznam<T> &seznam<T>::operator=(const seznam<T> & s)
{
	vyprazdni();
	if (!s.prazdny())
	{
		operator+=(s);
	}
	cout << "operator = &" << endl;
	return *this;
}

template<typename T>
seznam<T> &seznam<T>::operator=(seznam<T> && s)
{
	//vyprazdni();
	tunel(s);
	cout << "operator = &&" << endl;
	return *this;
}


template<typename T>
void seznam<T>::vyprazdni()
{
	if (prazdny()) return;
	for (prvek* pom = hlava; !prazdny(); pom = hlava)
	{
		hlava = pom->dalsi();
		delete pom;
	}
	aktualni = zarazka;
}

template<typename T>
bool seznam<T>::prazdny() const
{
	return hlava == zarazka;
}

template<typename T>
seznam<T>::~seznam()
{
	vyprazdni();
	delete hlava;
	cout << "destruktor" << endl;
}


template<typename T>
class seznam<T>::prvek
{
	T hodnota;
	prvek *_dalsi;
public:
	prvek(T hodnota, prvek *dalsi = nullptr)
		: hodnota(hodnota), _dalsi(dalsi)
	{}
	T &data();
	void dalsi(prvek * dal);
	prvek *dalsi();
};

template<typename T>
typename seznam<T>::prvek
*seznam<T>::prvek::dalsi()
{
	return _dalsi;
}

template<typename T>
T &seznam<T>::prvek::data()
{
	return hodnota;
}

template<typename T>
void seznam<T>::prvek::dalsi(typename seznam<T>::prvek *dal)
{
	_dalsi = dal;
}

#endif